from conML.controller import get_settings
from conML.controller import knowledge_searcher
from conML.controller import load_settings
from conML.controller import load_knowledge
from conML.controller import construction
from conML.controller import feature_selection
from conML.controller import reconstruction
from conML.controller import deconstruction


def _load_default_settings():
    from os.path import join, expanduser, exists
    path = join(expanduser('~'), ".conML", "settings.ini")
    if exists(path): load_settings(path)


def _setup_loggers():
    from logging import addLevelName, Logger, config
    from os.path import join, abspath
    from inspect import getsourcefile

    protocol_level = 55
    addLevelName(protocol_level, "PROTOCOL")

    def protocol(self, message, *args, **kwargs):
        if self.isEnabledFor(protocol_level):
            self._log(protocol_level, message, args, **kwargs)
    Logger.protocol = protocol
    logging_config = join(abspath(getsourcefile(lambda: None))[:-11],
                          "static/logging.ini")
    config.fileConfig(logging_config)


_load_default_settings()
_setup_loggers()



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Define the 'Complete' deconstruction.

"""


from conML.domain.deconstruction.abstract import Abstract
from conML.shared.logger import DeconstructionInfo


__all__ = (
    "CompleteDeconstructor",
)


class CompleteDeconstructor(Abstract):
    """Class performs the 'Complete' deconstruction process.

    Args:
        *args:
        cluster_finder (ClusterFinder): Find cluster within time stamps.
        general_settings (GeneralSettings): Controlls behavior of the
            deconstruction process.
        block_processing_settings (BlockProcessingSettings): Controlls
            the behavior of the deconstruction process.

    """
    RELATION = ("complete", )
    MIN_N_FEATURE = 2

    def __init__(self, *args, cluster_finder, general_settings,
                 block_processing_settings):
        super().__init__(*args)
        self.cluster_finder = cluster_finder
        self.general_settings = general_settings
        self.block_processing_settings = block_processing_settings

    def deconstruct(self, relative, log):
        """Run the 'Complete' deconstruction.

        Args:
            relative (PragmaticMachineLearningModel): Relative model.
            log (DeconstructionInfo): Logging information.

        Returns:
            DeconstructionInfo:
                Logging information.

        """
        if self.feature_constraint():
            meta = self.model_fusion(relative)
            learnblock = self.block.same_features_fusion(self.relative_block)
            return self.reconstructing(meta, learnblock, relative, log)

        elif self.time_stamp_constraint():
            if self.reliability_constraint():
                meta = self.model_fusion(relative)
                learnblock = self.build_block_from_diff_features()
                return self.reconstructing(meta, learnblock, relative, log)
            else:
                modification = self.model_disposal(self.pragmatic, relative)
                log.inserted.extend(modification.inserted)
                log.deleted.extend(modification.deleted)
                log.state = DeconstructionInfo.State.FAILED
                return log

        else:
            self.knowledge.insert(self.pragmatic)
            log.state = DeconstructionInfo.State.FAILED
            log.inserted.append(self.pragmatic)
            return log

    def feature_constraint(self):
        """Check if learnblocks of the recently reconstructed model and
        relative have enough identical features.

        Returns:
            bool:
                True, if learnblocks have enough identical features.

        """
        identical_features = len(
            set(self.block.columns(effective=True)).intersection(
                set(self.relative_block.columns(effective=True))
            )
        )
        return identical_features >= self.MIN_N_FEATURE

    def reconstructing(self, meta, learnblock, relative, log):
        """Reconstruct the new combined pragmatic machine learning model.

        Args:
            meta (MetaData): Meta data of combined pragmatic machine learning
                model.
            learnblock (PandasBlock): Valid learnblock.
            relative (PragmaticMachineLearningModel): Relative model.
            log (DeconstructionInfo): Logging information.

        Returns:
            DeconstructionInfo:
                Holds information about complete deconstruction.

        """
        if (len(learnblock.columns(effective=True)) < 2 or
                learnblock.n_rows() == 0):
            log.state = DeconstructionInfo.State.FAILED
            modification = self.model_disposal(self.pragmatic, relative)
            log.inserted.extend(modification.inserted)
            log.deleted.extend(modification.deleted)
            return log

        _, pragmatics, _ = self.reconstructor.reconstruct(
            self.level, learnblock, which_models=None,
            meta=meta, all_m=True)

        if pragmatics:
            winners = {m.accuracy: m for m in pragmatics}
            winner = winners[max(winners.keys())]
            log.reconstructed_info.append(winner)
            deleted = self.knowledge.remove_dependent(relative)
            self.knowledge.insert(winner)
            log.inserted.append(winner)
            log.deleted.extend(deleted)
            log.state = DeconstructionInfo.State.DECONSTRUCTED
            return log

        else:
            self.model_differentiation(learnblock, relative, log)

    def model_differentiation(self, block, relative_model, log):
        """Try to find sub models within the failed reconstructed model during
        complete deconstruction.

        Args:
            block (PandasBlock): Valid learnblock.
            relative_model (PragmaticMachineLearningModel): Relative model.
            log (DeconstructionInfo): Logging information.

        Returns:
            DeconstructionInfo

        """
        time_column = block.get_column("T")
        clusters = list(
            self.cluster_finder.find_valid_learnblock_cluster(
                time_column
             )
        )

        for time_values in clusters:
            learnblock = block.new_block_from(time_values, time_column=True)

            self.push_ids_to_reconstructor(
                len(self.reconstructor.ml_models) * len(clusters)
            )
            _, pragmatics, _ = self.reconstructor.reconstruct(
                self.level, learnblock, which_models=None,
                meta=None, all_m=True)
            if pragmatics:
                winners = {m.accuracy: m for m in pragmatics}
                winner = winners[max(winners.keys())]
                log.reconstructed_info.append(winner)
                log.inserted.append(winner)
                self.knowledge.insert(winner)

        if log.inserted:
            log.state = DeconstructionInfo.State.DECONSTRUCTED
            return log

        else:
            deleted = self.knowledge.remove_dependent(relative_model)
            log.deleted.extend(deleted)
            log.state = DeconstructionInfo.State.DECONSTRUCTED
            return log

    def time_stamp_constraint(self):
        """Check if recently reconstructed model has enough identical
        timestamps as his relative.

        Returns:
            bool:
                True, if the timestamp constraint is satisfied else False.

        """
        overlapped = self.block.get_overlapping_on_timestamps(
            self.relative_block)
        if (overlapped.n_rows() >=
                self.block_processing_settings.learn_block_minimum):
            return True

        else:
            return False

    def reliability_constraint(self):
        """Check if learnlock labels of recently reconstructed model satisfy
        the inter-reliability constaint with the relative.

        Returns:
            bool:
                True, if learnblocks labels of both models have high enough
                inter-reliability.

        """
        overlapped = self.block.get_overlapping_on_timestamps(self.relative_block)
        overlapped.drop_columns(
            list(set(overlapped.columns()).difference({"Z", "Z_other"}))
        )
        reliability = self.calculate_reliability(
            overlapped.get_column("Z"),
            overlapped.get_column("Z_other"))
        return reliability >= self.settings.min_reliability

    def build_block_from_diff_features(self):
        """Build learnblock on the basis of different features from
        recently reconstructed model and his relative.

        Returns:
            PandasBlock

        """
        overlapped = self.block.get_overlapping_on_timestamps(self.relative_block)
        if (overlapped.n_columns() == 0 or overlapped.n_rows() == 0 or
                len(overlapped.columns(effective=True)) < 2):
            return overlapped

        dont_drop = (
            set(self.block.columns(effective=True)).symmetric_difference(
                set(self.relative_block.columns(effective=True)))
        )

        if self.settings.deconst_strategy == "conservative":
            z_column = "Z"

        elif self.settings.deconst_strategy == "integrative":
            z_column = "Z_other"

        else:
            # self.settings.deconst_mode == "oppurtunistic":
            if self.block.n_rows() > self.relative_block.n_rows():
                z_column = "Z"
            else:
                z_column = "Z_other"

        overlapped.drop_columns(
            list(
                set(overlapped.columns()).difference(
                    set(dont_drop).union(
                        {"T", "Sigma", z_column}))
            )
        )
        overlapped.rename_column(z_column, "Z")
        overlapped.reorder_columns(list(dont_drop) + ["T", "Sigma", "Z"])
        overlapped.origin = tuple(set(self.block.origin + self.relative_block.origin))

        return overlapped



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the construction process.

"""

from copy import deepcopy
from functools import partial

from conML.shared.logger import ConstructionInfo


def return_model(m):
    """Used to monkeypatch the machine_learning_models() method
    of a Constructor object.

    Args:
        m (ConstructionClusteringMLModel): A valid instantiated object.

    Yield:
        ConstructionClusteringMLModel:
            The only model used by the patched constructor.

    """
    yield m


class Constructor:
    """Encapsulates the construction procedure.


    Notes:
        Only conceptual mode is currently supported.

    Args:
        mode (str):
            Either 'conceptual' or 'procedural'.
        ml_models (list):
            A list with instantiated ConceptualClusteringMLModels.
        settings (ConstructionSettings):
            Instantiated object controlls the behavior of the Constructor.

    """
    def __init__(self, mode, ml_models, settings):
        self.mode = mode
        self.ml_models = ml_models
        self.settings = settings

    def __str__(self):
        return "\n".join([
            "{:20}: {}".format("Mode", str(self.mode)),
            "{:20}: {}".format("ML Models", [
                str(m) for m in self.ml_models]),
            str(self.settings),
            ]
        )

    def __repr__(self):
        return ", ".join([
            "{}={}".format("mode", str(self.mode)),
            "{}".format(repr(self.settings)),
            "{}={}".format("ml_models", repr(self.ml_models))
            ]
        )

    def construct(self, learnblock):
        """Main method for constructing a given learnblock.

        Args:
            learnblock (PandasBlock):
                Valid learnblock.

        Yields:
            tuple:
                A ConstructionInfo object containing information about
                the currently executed construction process regarding
                only one unsupervised machine learning model, and the
                the constructed learnblock.

        """
        if learnblock.labeled():
            yield (
                ConstructionInfo(
                    state=ConstructionInfo.State.PASSED,
                    learnblock=learnblock
                ),
                learnblock
            )

        else:
            for model in self.machine_learning_models():
                model.train(learnblock)

                if not self.valid_trained(model, learnblock.n_rows()):
                    yield (
                        ConstructionInfo(
                            state=ConstructionInfo.State.DISCARDED,
                            subject=model.abbreviation,
                            n_cluster=model.found_cluster
                        ),
                        None
                    )

                else:
                    labeled_learnblock = self.label_block(learnblock, model)
                    yield (
                        ConstructionInfo(
                            state=ConstructionInfo.State.CONSTRUCTED,
                            subject=model.abbreviation,
                            n_cluster=model.found_cluster,
                            learnblock=labeled_learnblock
                        ),
                        labeled_learnblock
                    )

    def machine_learning_models(self):
        """Yields valid configurations of unsupervised machine learning models
        regarding their cluster specificiation.

        Yields:
            ConstructionClusteringMLModel

        """
        for ml_model in self.ml_models:
            if ml_model.cluster_specific:
                for n_cluster in range(2, self.settings.max_categories+1):
                    yield ml_model.new_model_depending_on_cluster(n_cluster)
            else:
                yield ml_model.new_model()

    def valid_trained(self, model, n_rows):
        """Validate the trained unsupervised machine learning model.

        Args:
            model (ConstructionClusteringMLModel):
                Trained model.
            n_rows:
                Number of samples in learnblock, which should be constructed.

        Returns:
            bool:
                True, if the model is valid else False.

        """
        if model.found_cluster <= 1:
            return False

        for cluster, size in model.cluster_sizes.items():
            if size < self.settings.min_category_size*n_rows:
                return False

        return True

    def label_block(self, block, trained_model):
        """Assign the labels, origin, subject, sigma to the constructed
        learnblock.


        Args:
            block (PandasBlock):
                Successfully constructed learnblock.
            trained_model (ConstructionClusteringMLModel):
                Unsupervised machine learning model, which successfully
                constructed the given learnblock.

        Returns:
            PandasBlock:
                Labeled learnblock.

        """
        copied_block = deepcopy(block)
        labels = trained_model.get_labels()
        copied_block.set_column("Z", labels)
        copied_block.n_cluster = trained_model.found_cluster
        copied_block.origin = block.origin
        subject = "{}{:02}".format(trained_model.abbreviation,
                                   trained_model.found_cluster)
        copied_block.set_column("Sigma", subject)
        copied_block.subject = subject
        return copied_block

    def serial_constructor_per_combination(self):
        """Generate many constructor objects, which only using one configuraton
        regarding their cluster specification.

        Yields:
            Constructor

        """
        for model in self.machine_learning_models():
            constructor = Constructor(self.mode, self.ml_models, self.settings)
            constructor.machine_learning_models = partial(return_model, model)
            yield constructor

    @property
    def min_category_size(self):
        return self.settings.min_category_size

    @min_category_size.setter
    def min_category_size(self, value):
        self.settings.min_category_size = value

    @property
    def max_categories(self):
        return self.settings.max_categories

    @max_categories.setter
    def max_categories(self, value):
        self.settings.max_categories = value



from configparser import ConfigParser
from os import mkdir
from os.path import expanduser, join, exists
import pathlib
from setuptools import find_packages, setup, Extension


HERE = pathlib.Path(__file__).parent
README = (HERE/"README.txt").read_text()

USER_HOME_DIR = expanduser("~")
CML_DIR_NAME = ".conML"
DEFAULT_LEARN_DIR = "data"
CONFIG_FILE_NAME = "settings.ini"
DEFAULT_TRAINING_SET = "toyset.csv"
KNOWLEDGE_DIR = "knowledge"

CML_CONFIG_DIR = join(USER_HOME_DIR, CML_DIR_NAME)
CML_DEFAULT_LEARN_DIR = join(CML_CONFIG_DIR, DEFAULT_LEARN_DIR)
CONFIG_FILE_PATH = join(CML_CONFIG_DIR, CONFIG_FILE_NAME)
DEFAULT_TRAINING_SET_PATH = join(CML_DEFAULT_LEARN_DIR, DEFAULT_TRAINING_SET)
KNOWLEDGE_DIR_PATH = join(CML_CONFIG_DIR, KNOWLEDGE_DIR)


def create_cml_config_dir():
    mkdir(CML_CONFIG_DIR)


def create_data_dir():
    mkdir(CML_DEFAULT_LEARN_DIR)


def insert_default_settings_into_config_file():
    config = ConfigParser()
    config.read(CONFIG_FILE_PATH)
    config["GENERAL"]["input_file"] = DEFAULT_TRAINING_SET_PATH
    config["GENERAL"]["learn_dir"] = CML_DEFAULT_LEARN_DIR
    config["GENERAL"]["knowledge_dir"] = KNOWLEDGE_DIR_PATH
    write_configs(config)


def write_configs(config):
    with open(CONFIG_FILE_PATH, 'w') as configfile:
        config.write(configfile)


if not exists(CML_CONFIG_DIR):
    create_cml_config_dir()


setup(
    name = "conML",
    packages = find_packages(),
    version = 0.31,
    licence="MIT",
    description="constructivist machine learning",
    long_description=README,
    long_description_content_type="text/plain",
    data_files=[(CML_CONFIG_DIR, ["conML/static/settings.ini"]),
                (CML_DEFAULT_LEARN_DIR, ["conML/static/toyset.csv"]),
                ("conML/static", ["conML/static/logging.ini"])],
    author="Dmitrij Denisenko",
    install_requires = [
        "numpy",
        "krippendorff",
        "scikit-learn",
        "scipy",
        "pandas"
    ]
)



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the queries for the framework.

"""


from copy import copy, deepcopy
from collections import defaultdict, namedtuple
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler
import multiprocessing
from queue import Empty
import os


from conML.domain.knowledge import KnowledgeDatabase, RelativeFinder
from conML.domain.preprocessing import Source, LearnblockIdentifier
from conML.shared.errors import ModeError
from conML.shared.logger import JsonContainer, HighLevelLearnblockInfo
from conML.shared.parameter import PROTOCOL_LEVEL
from conML.usecases.usecase import Usecase


class ParallelRunner:
    """Responsible for performing a single construction and reconstruction
    in a separate process.

    Args:
        constructor (Constructor): Constructs the learnblock.
        selector (FeatureSelector): Selects the most important features.
        reconstructor (Reconstructor): Reconstructs the learnblock.
        log_queue (Queue): Stores the logging information of the
            performed construction, selection and reconstrution process.

    """

    def __init__(self, constructor, selector, reconstructor, log_queue):
        self.constructor = constructor
        self.selector = selector
        self.reconstructor = reconstructor
        self.log_queue = log_queue
        self.result = defaultdict(list)

    def run(self, level, block):
        """Run the construction and reconstruction on the given block.

        Args:
            level (int):
                Level on which the construction and reconstruction should be
                performed.
            block (PandasBlock): Learnblock.

        Returns:
            defaultdict:
                Maps the reliability to PragmaticMachineLearningModel.

        """
        it = iter(self.constructor.construct(block))
        con_info, labeled_lb = next(it)
        self.log_queue.put(con_info)

        if not labeled_lb:
            return

        sel_info, reduced_lb = self.selector.select(labeled_lb)
        self.log_queue.put(sel_info)

        if not reduced_lb:
            return

        reliability, pragmatics, rec_info = self.reconstructor.reconstruct(
            level, reduced_lb, which_models=None, meta=None)
        self.log_queue.put(rec_info)

        self.result[reliability].extend(pragmatics)
        return self.result


def map_onto(args):
    """Executed by every pool worker.

    Args:
        args (tuple):

    Returns:

    """
    parallel_runner, level, block = args
    return parallel_runner.run(level, block)


class KnowledgeSearcher:
    """Responsible for coordinating the construction, reconstruction and
    deconstruction.

    Args:
        source (Source): Converts given pandas.DataFrames to PandasBlocks.
        identifier (LearnblockIdentifier): Identifies learnblocks within
            PandasBlocks.
        constructor (Constructor): Construct the learnblocks.
        reconstructor (Reconstructor): Reconstructs the learnblocks.
        selector (FeatureSelctor): Selects subset of learnblock features.
        deconstructor (Deconstructor): Deconstruct the reconstructed
            PragmaticMachineLearningModel.
        knowledge_database (KnowledgeDatabase): Database to store models.
        stdout (bool): Indicator if the process should logged to stdout.
        parallel (bool): Indicator if the learning process should utilize
            many processes.
        n_procs (int): Number of processes to utilize.

    """
    def __init__(self,
                 source,
                 identifier,
                 constructor,
                 reconstructor,
                 selector,
                 deconstructor,
                 knowledge_database,
                 stdout,
                 n_procs):
        self.source = source
        self.identifier = identifier
        self.constructor = constructor
        self.reconstructor = reconstructor
        self.selector = selector
        self.deconstructor = deconstructor
        self.knowledge_database = knowledge_database
        self.stdout = stdout
        self.n_procs = n_procs

        self.pool = None
        self.logger = None
        self.json_container = None
        self.log_destination = None
        self.tsigma_beamer = deconstructor.tsigma_beamer

    def __enter__(self):
        self.log_destination = self.__get_log_destination()
        self.logger = self.__get_logger()
        self.json_container = JsonContainer()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.json_container.write_down(self.log_destination)

        del self.logger
        del self.json_container
        del self.log_destination
        del self.tsigma_beamer

    def __get_log_destination(self):
        """Create folder to store logs.

        Returns:
            str:
                Logging destination.

        """
        home_folder_path = os.path.expanduser("~")
        cml_folder = ".conML"
        current_time = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
        log_dir = os.path.join(home_folder_path, cml_folder, current_time)
        os.mkdir(log_dir)
        return log_dir

    def search(self, block):
        """Start the knowledge search process.

        Args:
            block (pandas.DataFrame): Raw block.

        Returns:
            tuple:
                KnowledgeDatabase, pandas.DataFrame

        """
        if self.n_procs > 1:
            try:
                n_procs = self.n_procs
                self.pool = multiprocessing.Pool(n_procs, maxtasksperchild=1)
                return self.search_parallel(block)

            finally:
                if self.pool:
                    self.pool.close()
                    self.pool.join()
                    del self.pool

        else:
            return self.search_serial(block)

    def search_parallel(self, block):
        """Perform the knowledge search process with the help of many processes.

        Args:
            block (pandas.DataFrame): Raw data block.

        Returns:
            tuple:
                KnowledgeDatabase, pandas.DataFrame

        """
        halde = None
        for gen in self._generate_learnblock(block):
            if halde is None:
                halde = gen.halde

            if gen.lb is None:
                break

            candidates = defaultdict(list)

            level = gen.level
            lb = gen.lb

            log_queues_list, runner_list = self.get_parallel_runner(level)
            args = [(runner, level, lb) for runner in runner_list]

            reliability_dicts = self.pool.map(map_onto, args)

            self.log_parallel_version(log_queues_list)
            for rel_dict in reliability_dicts:
                if rel_dict is not None:
                    reliability, pragmatics = rel_dict.popitem()
                    if reliability == 0.0:
                        continue

                    candidates[reliability].extend(pragmatics)

            win_info, winner = self.reconstructor.select_winner(candidates)
            self.log(win_info)

            if winner:
                for dec_infs in self.deconstructor.deconstruct(
                        gen.level, winner, self.knowledge_database):
                    self.log(dec_infs)

        return deepcopy(self.knowledge_database), halde

    def get_parallel_runner(self, level):
        """Build n ParallelRunners.

        n = number of unsupervised algorithms * cluster specification

        Args:
            level (int): Level on which the ParallelRunner should
                perform the construction and reconstruction.

        Returns:
            tuple:
                Queue, ParallelRunner

        """
        runners = []
        log_queues = []
        manager = multiprocessing.Manager()
        identifier_queue = manager.Queue()

        for constructor in self.constructor.serial_constructor_per_combination():
            reconstructor = copy(self.reconstructor)
            reconstructor.identifier_queue = identifier_queue

            selector = self.selector

            log_queue = manager.Queue()
            log_queues.append(log_queue)

            runner = ParallelRunner(
                constructor,
                selector,
                reconstructor,
                log_queue,
            )
            runners.append(runner)

        n_ids = len(runners) * len(self.reconstructor.ml_models)
        for _ in range(n_ids):
            identifier = self.knowledge_database.get_next_identifier(level)
            identifier_queue.put(identifier)

        return log_queues, runners

    def log_parallel_version(self, log_queues):
        """Empty the log queues and protocol the information that was
        stored in logs.

        Args:
            log_queues (list): Log queues of the ParallelRunners.

        """
        for log_queue in log_queues:
            while not log_queue.empty():
                try:
                    info = log_queue.get(False)
                    self.log(info)
                except Empty:
                    continue

    def search_serial(self, block):
        """Perform the the knowledge search process.

        Args:
            block (pandas.DataFrame): Raw block.

        Returns:
            tuple:
                KnowledgeDataFrame, pandas.DataFrame

        """
        halde = None
        for gen in self._generate_learnblock(block):
            if halde is None:
                halde = gen.halde

            if gen.lb is None:
                break

            candidates = defaultdict(list)

            for con_info, labeled_lb in self.constructor.construct(gen.lb):
                self.log(con_info)

                if labeled_lb is None:
                    continue

                sel_info, reduced_lb = self.selector.select(labeled_lb)
                self.log(sel_info)

                if reduced_lb is None:
                    continue

                self.calculate_needed_ids(gen.level, self.reconstructor)
                rel, interims, rec_infs = self.reconstructor.reconstruct(
                    gen.level, reduced_lb, which_models=None, meta=None
                )
                self.log(rec_infs)

                if rel == 0.0:
                    continue

                candidates[rel].extend(interims)

            win_inf, winner = self.reconstructor.select_winner(candidates)
            self.log(win_inf)

            if winner:
                for dec_infs in self.deconstructor.deconstruct(
                        gen.level, winner, self.knowledge_database):
                    self.log(dec_infs)

        return deepcopy(self.knowledge_database), halde

    def log(self, log_info):
        """Log the information that happened during the knowledge search
        process.

        Args:
            log_info (LogMeMixin): Holds logging information.

        """
        try:
            for info in log_info:
                info.log_me(self.logger, self.stdout, self.json_container)

        except TypeError:
            log_info.log_me(self.logger, self.stdout, self.json_container)

    def _generate_learnblock(self, block):
        """Yield learnblocks for the knowledge search process.

        Args:
            block (pandas.DataFrame): Raw block.

        Yield:
            PandasBlock:
                Valid learnblock.

        """
        level = 1
        Generated = namedtuple("Generated", ("level", "lb", "halde"))

        proc_info, pandas_block = self.source.process(block)
        proc_info.level = level
        self.log(proc_info)

        idn_info, learnblock, halde = self.identifier.identify(pandas_block)
        idn_info.level = level
        self.log(idn_info)
        if learnblock is None:
            yield Generated(level, None, halde)

        self.source.append_to_knowledge(learnblock)
        yield Generated(level, learnblock, halde)

        while self.tsigma_beamer:
            level, learnblock = self.tsigma_beamer.pop()
            if level and learnblock:
                self.log(HighLevelLearnblockInfo(learnblock, level))
                yield Generated(level, learnblock, None)

    def calculate_needed_ids(self, level, reconstructor):
        """Push enough new identifier into identifier_queue of the given
        reconstructor.

        Args:
            level (int): Level of KnowledgeDatabase.
            reconstructor (Reconstructor): Where to push the identifiers.

        """
        n_ids = len(reconstructor.ml_models)
        for _ in range(n_ids):
            identifier = self.knowledge_database.get_next_identifier(level)
            reconstructor.identifier_queue.put(identifier)

    def __get_logger(self):
        """Return the logger object.

        Returns:
            logging.RootLogger

        """
        iteration_log_file = os.path.join(self.log_destination, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")

        # Remove loggers from other sessions
        if iteration_logger.handlers:
            iteration_logger.handlers.clear()
        iteration_logger.addHandler(iteration_handler)

        if self.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)

        return iteration_logger


class KnowledgeSearcherUsecase(Usecase):
    def process(self, request):
        """Process the request to create a KnowledgeSearcher.

        Args:
            request (KnowledgeSearchingRequest): Namedtuple with all essential
                ingredients.

        Raises:
            ModeError:
                Inconsistent Reconstructor and Constructor modes.

        Returns:
            KnowledgeSearcher

        """
        if not self.valid_modes(request):
            raise ModeError(
                "Inconsistent constructor and reconstructor modes."
                "Returns KnowledgeSearcher object by consistent modes."
                "Make sure the modes of constructor and reconstructor are the "
                "same."
            )

        return KnowledgeSearcher(**self.build_parts(request))

    def valid_modes(self, request):
        """Check if the given Constructor and Reconstructor as well as
        Deconstructor have the same mode.

        Args:
            request (KnowledgeSearchingRequest): Namedtuple with all essential
                ingredients.

        Returns:
            bool:
                True, if all mode are consistent.

        """
        return request.constructor.mode == request.reconstructor.mode

    def build_parts(self, request):
        """Build all essential components for the knowledge search process.


        Args:
            request (KnowledgeSearchingRequest): Namedtuple with all essential
                ingredients.

        Returns:
            dict:
                Map attribute name to the acctually attribute.

        """
        knowledge = self.create_database(request)
        identifier = LearnblockIdentifier(
            request.block_processing_settings,
            request.density_estimator,
            request.ckmeans
        )
        source = Source(knowledge, request.df_converter)
        return {
            "source": source,
            "identifier": identifier,
            "constructor": request.constructor,
            "selector": request.feature_selector,
            "reconstructor": request.reconstructor,
            "deconstructor": request.deconstructor,
            "knowledge_database": knowledge,
            "stdout": request.stdout,
            "n_procs": request.n_procs
        }

    def create_database(self, request):
        """Create a new KnowledgeDatabase if the given request does not
        contain any database.

        Args:
            request (KnowledgeSearchingRequest): Namedtuple with all essential
                ingredients.

        Returns:
            KnowledgeDatabase

        """
        if request.knowledge is not None:
            knowledge = request.knowledge

        else:
            knowledge = KnowledgeDatabase(
                request.general_settings.highest_level,
                request.build_block_from_rows
            )
            knowledge.observer.append(
                RelativeFinder(request.general_settings.highest_level)
            )

        return knowledge



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


PROTOCOL_LEVEL = 55


def value_error_string(name, to):
    """Teample for ValueErrors.

    Args:
        name (str): Parameter name.
        to (str): Target data type.

    Returns:
        str

    """
    return (
        f"Invalid value for {name}. "
        f"Tried to convert string to {to}. "
        f"Make sure that the value is a valid {to} string."
    )


def type_error_string(name, type_):
    """Template for TypeErrors.

    Args:
        name (str): Parameter name.
        type_ (str): Type of the assigned value to the parameter.

    Returns:
        str

    """
    return (
        f"Invalid type for {name}. "
        f"Tried to redefine {name}. "
        f"Make sure that the value is of type {type_} or string."
    )


class HighestLevel:
    """Largest level of the knowledge database.

    """
    def __init__(self):
        self.highest_level = 4

    def __get__(self, instance, owner):
        return self.highest_level

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.highest_level = value
        elif isinstance(value, str):
            try:
                self.highest_level = int(value)
            except ValueError:
                raise ValueError(value_error_string("highest_level", "integer"))
        else:
            raise TypeError(type_error_string("highest_level", "int"))


class KernelBandwidth:
    """Controlls the bandwidth of the kernel used during density estimation.

    """
    def __init__(self):
        self.kernel_bandwidth = 0.2

    def __get__(self, instance, owner):
        return self.kernel_bandwidth

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.kernel_bandwidth = value
        elif isinstance(value, str):
            try:
                self.kernel_bandwidth = float(value)
            except ValueError:
                raise ValueError(
                    value_error_string("kernel_bandwidth", "float")
                )
        else:
            raise TypeError(type_error_string("kernel_bandwidth", "float"))


class LearnblockMinimum:
    """Minimum required number of samples per learning block. If learning
    blocks are identified within a drawn block that contain fewer samples,
    these learning blocks are discarded. This value must always be set, a
    default value is not implemented.

    """
    def __init__(self):
        self.learnblock_minimum = 500

    def __get__(self, instance, owner):
        return self.learnblock_minimum

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.learnblock_minimum = value

        elif isinstance(value, str):
            try:
                self.learnblock_minimum = int(value)
            except ValueError:
                raise ValueError(value_error_string("learnblock_minimum",
                                                    "integer"))
        else:
            raise TypeError(type_error_string("learnblock_minimum", "int"))


class MaxCategories:
    """Maximale kategoriale Komplexität.
    Bestimmt innerhalb der konzeptuellen Wissensdomäne die Anzahl der Cluster,
    die mittels Clustering-Verfahren bestimmt werden. Dieser Wert muss immer
    gesetzt werden, ein Default-Wert ist nicht implementiert.

    """
    MIN_VALUE = 2
    MAX_VALUE = 1000

    def __init__(self):
        self.max_categories = 2

    def __get__(self, instance, owner):
        return self.max_categories

    def __set__(self, instance, value):
        if isinstance(value, int):
            if self.valid_range(value):
                self.max_categories = value

        elif isinstance(value, str):
            try:
                int_value = int(value)
                if self.valid_range(int_value):
                    self.max_categories = int_value
            except ValueError:
                raise ValueError(value_error_string("max_categories",
                                                    "integer"))

        else:
            raise TypeError(type_error_string("max_categories", "int"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError(
                "Invalid value for max_categories. "
                "Tried to redefine max_categories. "
                "Make sure that the value is greater equal two.")
        return True


class MinCategorySize:
    """Minimum size of a category. Creates the required number within the
    conceptual domain of knowledge samples per cluster. If this is not used
    by all Clustering of a clustering is achieved, the clustering as a whole
    is discarded. Instead of an integer a decimal value between 0 and 1 is
    also given with which the sample number can be calculated from the
    total number of of the samples of the learning block is calculated. This
    value must always be set, a default value is not implemented.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.min_category_size = 0.1

    def __get__(self, instance, owner):
        return self.min_category_size

    def __set__(self, instance, value):
        if isinstance(value, float):
            if self.valid_range(value):
                self.min_category_size = value

        elif isinstance(value, str):
            try:
                float_value = float(value)
                if self.valid_range(float_value):
                    self.min_category_size = float(value)
            except ValueError:
                raise ValueError(value_error_string("min_category_size",
                                                    "float"))

        else:
            raise TypeError(type_error_string("min_category_size", "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for min_category_size. "
                             "Tried to redefine min_category_size. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class SigmaZetaCutoff:
    """Threshold value for the identification of SigmaZ-related Study blocks.
    Determines which density distribution of the Metadata T in a drawn block
    at least must be achieved in order to achieve a one-dimensional clustering.
    Areas that fall below this threshold are not considered.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.sigma_zeta_cutoff = 0.1

    def __get__(self, instance, owner):
        return self.sigma_zeta_cutoff

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.sigma_zeta_cutoff = value
        elif isinstance(value, str):
            try:
                self.sigma_zeta_cutoff = float(value)
            except ValueError:
                raise ValueError(value_error_string("sigma_zeta_cutoff",
                                                    "float"))
        else:
            raise TypeError(type_error_string("sigma_zeta_cutoff", "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for sigma_zeta_cutoff. "
                             "Tried to redefine sigma_zeta_cutoff. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class MaxModelTargets:
    def __init__(self):
        self.max_model_targets = 10

    def __get__(self, instance, owner):
        return self.max_model_targets

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.max_model_targets = value

        elif isinstance(value, str):
            try:
                self.max_model_targets = int(value)
            except Exception as error:
                print(error)


class MaxTargetError:
    def __init__(self):
        self.max_target_error = 0.8

    def __get__(self, instance, owner):
        return self.max_target_error

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.max_target_error = value

        elif isinstance(value, str):
            try:
                self.max_target_error = float(value)
            except TypeError as error:
                print(error)


class MaxFeatures:
    """Maximum allowable model complexity. With this value defines the
    maximum number of input values a model may have. From MaxFeatures+1
    input values a complexity reduction is carried out.

    """
    def __init__(self):
        self.max_features = 5

    def __get__(self, instance, owner):
        return self.max_features

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.max_features = value
        elif isinstance(value, str):
            try:
                self.max_features = int(value)
            except ValueError:
                raise ValueError(value_error_string("max_features",
                                                    "int"))
        else:
            raise TypeError(type_error_string("max_features", "int"))


class MaxFilterX:
    """Filter limit. This value controls when an embedded feature selection
    procedure and when an filter method is used. If a model contains more
    than MaxFilterX input values, a filter method is used; otherwise an embedded
    procedure. This value must always be set, a default value is not implemented

    """
    def __init__(self):
        self.max_filter_x = 10

    def __get__(self, instance, owner):
        return self.max_filter_x

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.max_filter_x = value
        elif isinstance(value, str):
            try:
                self.max_filter_x = int(value)
            except ValueError:
                raise ValueError(value_error_string("max_filter_x",
                                                    "int"))
        else:
            raise TypeError(type_error_string("max_filter_x", "int"))


class MaxFilterY:
    """Filter limit. This value controls when an embedded feature selection
    method and when a filtering method is used. If a model contains more
    as MaxFilterY samples, a filter method is used; otherwise an embedded
    procedure. This value must always be set, a default value is not
    implemented.

    """
    def __init__(self):
        self.max_filter_y = 1000

    def __get__(self, instance, owner):
        return self.max_filter_y

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.max_filter_y = value
        elif isinstance(value, str):
            try:
                self.max_filter_y = int(value)
            except ValueError:
                raise ValueError(value_error_string("max_filter_y",
                                                    "int"))
        else:
            raise TypeError(type_error_string("max_filter_y", "int"))


class MaxModelsReduction:
    """Forcing a maximum reduction of complexity.
    If this value is set to true, complexity reduction steps are performed until
    no more features can be removed. Otherwise the complexity reduction is
    aborted as soon as less than MaxFeatures input variables were achieved.

    """
    def __init__(self):
        self.max_models_reduction = False

    def __get__(self, instance, owner):
        return self.max_models_reduction

    def __set__(self, instance, value):
        if isinstance(value, bool):
            self.max_models_reduction = value
        elif isinstance(value, str):
            if value in ("no", "No", "false", "False"):
                self.max_models_reduction = False
            elif value in ("yes", "Yes", "true", "True"):
                self.max_models_reduction = True
            else:
                raise ValueError(value_error_string("max_models_reduction",
                                                    "bool"))
        else:
            raise TypeError(type_error_string("max_models_reduction", "bool"))


class MinTestAccuracy:
    """Accuracy check threshold. Defines within the conceptual domain of
    knowledge which Accuracy value models for predictions of classifications
    must at least be achieved. Will MinTestAccuracy is not reached, the model
    discarded.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.min_test_accuracy = 0.8

    def __get__(self, instance, owner):
        return self.min_test_accuracy

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.min_test_accuracy = value
        elif isinstance(value, str):
            try:
                self.min_test_accuracy = float(value)
            except ValueError:
                raise ValueError(value_error_string("min_test_accuracy",
                                                    "float"))
        else:
            raise TypeError(type_error_string("min_test_accuracy", "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for min_test_accuracy. "
                             "Tried to redefine min_test_accuracy. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class MaxTestErrorAvg:
    """Threshold value for the average prediction error. Defines the maximum
    average percentage deviation from the target value for regression
    predictions within the procedural knowledge domain. Is MaxTestErrorAvg is
    exceeded, the model discarded.

    """
    def __init__(self):
        self.max_test_error_avg = 10

    def __get__(self, instance, owner):
        return self.max_test_error_avg

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.max_test_error_avg = value
        elif isinstance(value, str):
            try:
                self.max_test_error_avg = float(value)
            except TypeError as error:
                print(error)


class MaxTestErrorMax:
    """Threshold value for the maximum prediction error.
    Defines within the procedural knowledge domain, the maximum percentage
    deviation from the target value for regression predictions that may be
    achieved. If MaxTestErrorAvg is exceeded, the model is discarded.

    """
    def __init__(self):
        self.max_test_error_max = 10

    def __get__(self, instance, owner):
        return self.max_test_error_max

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.max_test_error_max = value
        elif isinstance(value, str):
            try:
                self.max_test_error_max = float(value)
            except TypeError as error:
                print(error)


class MinBuildModels:
    """Minum of successful trained supervised machine learning models to
    continue the reconstruction process.

    """
    MIN_VALUE = 2

    def __init__(self):
        self.min_build_models = self.MIN_VALUE

    def __get__(self, instance, owner):
        return self.min_build_models

    def __set__(self, instance, value):
        if isinstance(value, int):
            self.min_build_models = value
        elif isinstance(value, str):
            try:
                self.min_build_models = int(value)
            except ValueError:
                raise ValueError(value_error_string("min_build_models",
                                                    "int"))
        else:
            raise TypeError(type_error_string("min_build_models", "int"))


class ReliabilitySample:
    """Sample size for reliability testing. Determines the Proportion of
    samples used to determine the reliability coefficient of a model. The
    specified value must be between 0 and 1.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.reliability_sample = 0.2

    def __get__(self, instance, owner):
        return self.reliability_sample

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.reliability_sample = value
        elif isinstance(value, str):
            try:
                self.reliability_sample = float(value)
            except ValueError:
                raise ValueError(value_error_string("reliability_sample",
                                                    "float"))
        else:
            raise TypeError(type_error_string("reliability_sample", "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for reliability_sample. "
                             "Tried to redefine reliability_sample. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class MinReliability:
    """Threshold value for reliability testing. Determines from which value
    of the interreliability coefficient a agreement is assumed.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.min_reliability = 0.1

    def __get__(self, instance, owner):
        return self.min_reliability

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.min_reliability = value
        elif isinstance(value, str):
            try:
                self.min_reliability = float(value)
            except ValueError:
                raise ValueError(value_error_string("min_reliability",
                                                    "float"))
        else:
            raise TypeError(type_error_string("min_reliability", "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for reliability_sample. "
                             "Tried to redefine reliability_sample. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class ReduceModelRedundancy:
    def __init__(self):
        self.reduce_model_redundancy = False

    def __get__(self, instance, owner):
        return self.reduce_model_redundancy

    def __set__(self, instance, value):
        if isinstance(value, bool):
            self.reduce_model_redundancy = value

        elif isinstance(value, str):
            if value in ("False", "no", "false"):
                self.reduce_model_redundancy = False
            elif value in ("True", "yes", "true"):
                self.reduce_model_redundancy = True
            else:
                raise ValueError(value_error_string("reduce_model_redundancy",
                                                    "bool"))
        else:
            raise TypeError(type_error_string("reduce_model_redundancy",
                                              "bool"))


class DeconstStrategy:
    """Dealing with new models. Determines whether new successfully
    deconstructed models are included in the knowledge domain.

    """
    def __init__(self):
        self.deconst_strategy = "conservative"

    def __get__(self, instance, owner):
        return self.deconst_strategy

    def __set__(self, instance, value):
        if value in ("conservative", "integrative", "oppurtunistic"):
            self.deconst_strategy = value
        else:
            raise ValueError(
                f"Invalid value for deconst_strategy. "
                f"Tried to redefine deconst_strategy. "
                f"Make sure that the value is either 'conservative', "
                f"'integrative', 'oppurtunistic'."
            )


class DeconstMode:
    """Termination condition. The deconstruction process is either aborted
    as soon as a complete, a SigmaZ, a TZ or a TSigma deconstruction was
    successful (minimal) or only when no more pragmatically related models
    can be identified (full).

    """
    def __init__(self):
        self.deconst_mode = "minimal"

    def __get__(self, instance, owner):
        return self.deconst_mode

    def __set__(self, instance, value):
        if value in ("minimal", "full"):
            self.deconst_mode = value
        else:
            raise ValueError(
                f"Invalid valid for deconst_mode. " 
                f"Tried to redefine deconst_mode. "
                f"Make sure that the value is either 'full' or 'minimal'."
            )


class DeconstMaxDistanceT:
    """Temporal expansion tolerance. Defines the maximum size of the temporal
    gap between SigmaZ-related models for expansion to be tested. If this value
    is set to 1, the check is always performed. If it is set to 0, the check
    only checks if there is an overlap. Otherwise, the temporal gap must not
    exceed the size of the part of the total span defined by the set value.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.deconst_max_distance_t = 0.1

    def __get__(self, instance, owner):
        return self.deconst_max_distance_t

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.deconst_max_distance_t = value

        elif isinstance(value, str):
            try:
                self.deconst_max_distance_t = float(value)
            except ValueError:
                raise ValueError(value_error_string("deconst_max_distance_t",
                                                    "float"))

        else:
            raise TypeError(type_error_string("deconst_max_distance_t",
                                              "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for deconst_max_distance_t. "
                             "Tried to redefine deconst_max_distance_t. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class DeconstFullTolerance:
    """Temporal tolerance with complete kinship. Defines the proportion by
    which the time interval T can be extended to constitute a relationship.

    """
    MIN_VALUE = 0.0
    MAX_VALUE = 1.0

    def __init__(self):
        self.deconst_full_tolerance = 0.1

    def __get__(self, instance, owner):
        return self.deconst_full_tolerance

    def __set__(self, instance, value):
        if isinstance(value, float):
            self.deconst_full_tolerance = value

        elif isinstance(value, str):
            try:
                self.deconst_full_tolerance = float(value)

            except ValueError:
                raise ValueError(value_error_string("deconst_full_tolerance",
                                                    "float"))

        else:
            raise TypeError(type_error_string("deconst_full_tolerance",
                                              "float"))

    def valid_range(self, value):
        if not self.MIN_VALUE <= value <= self.MAX_VALUE:
            raise ValueError("Invalid value for deconst_full_tolerance. "
                             "Tried to redefine deconst_full_tolerance. "
                             "Make sure that the value is between 0.0 and 1.0.")
        return True


class ForceTimeExpansion:
    """Association of SigmaZ-related models. If True, SigmaZ-related models
    are treated preferentially or unified.

    """
    def __init__(self):
        self.force_time_expansion = True

    def __get__(self, instance, owner):
        return self.force_time_expansion

    def __set__(self, instance, value):
        if isinstance(value, bool):
            self.force_time_expansion = value

        elif isinstance(value, str):
            if value in ("False", "no"):
                self.force_time_expansion = False
            elif value in ("True", "yes"):
                self.force_time_expansion = True
            else:
                raise ValueError(value_error_string("force_time_expansion",
                                                    "bool"))
        else:
            raise TypeError("Invalid value for force_time_expansion. "
                            "Tried to redefine force_time_expansion. "
                            "Make sure that the value is a bool.")


class AllowWeakReliability:
    def __init__(self):
        self.allow_weak_reliability = False

    def __get__(self, instance, owner):
        return self.allow_weak_reliability

    def __set__(self, instance, value):
        if isinstance(value, bool):
            self.allow_weak_reliability = value

        elif isinstance(value, str):
            if value in ("False", "no"):
                self.allow_weak_reliability = False
            elif value in ("True", "yes"):
                self.allow_weak_reliability = True
            else:
                raise ValueError(value_error_string("allow_weak_reliability",
                                                    "bool"))

        else:
            raise TypeError("Invalid value for the allow_weak_reliability. ",
                            "Tried to redefine allow_weak_reliability. "
                            "Make sure that the value is a bool.")



#  C O N S T R U C T I V I S T__ __
#   _____ ____   ____   /  |/  // /
#  / ___// __ \ / __ \ / /|_/ // /
# / /__ / /_/ // / / // /  / // /___
# \___/ \____//_/ /_//_/  /_//_____/
#  M A C H I N E   L E A R N I N G
#
#
# A Project by
# Thomas Schmid | UNIVERSITÄT LEIPZIG
# www.constructivist.ml
#
# Code Author: Dmitrij Denisenko
# Licence: MIT


"""Defines the wrapper for calculating the inter-reliabiilty coefficient.

"""


import krippendorff


def krippendorff_alpha(voter):
    """Wrapper for krippendorff - alpha interreliability.

   Calculates the consensus between voters.

    Args:
        voter (list): Multidimensional array.

    Returns:
        float:
            Interreliability between voters.
            1.0 perfect consensus between voter.
            0.0 no consensus at all.
            -1.0 contradiction between voter.


    """
    val = krippendorff.alpha(voter)
    return round(val, 3)



"""Defines the reconstruction process.

"""


from collections import namedtuple
from itertools import zip_longest
from multiprocessing import Manager

from conML.shared.logger import ReconstructionInfo
from conML.domain.reconstruction.selection import WinnerSelector


InterimPragmatic = namedtuple("InterimPragmatic", "model, predictions")


class Metadata:
    """Responsible for storing the meta data of a pragmatic machine
    learning model.

    Args:
        knowledge_domain (str): Name of the knowlede domain.
        knowledge_level (int): Level on which model is stored.
        identifier (int): Unique model identifier.
        learnblock_indices (list): Indices of used learnblock.
        learnblock_features (list): Name of features used in learnblock.
        learnblock_labels (list): List of labels used in the learnblock.
        t_min (int): The smallest timestamp in used in learnblock.
        t_max (int): The biggest timestamp in used in learnblock.
        subject (tuple): Contains abbreviations of supervised
            machine learning models.
        aim (tuple):
            Tuple of strings in format
            'knowledge_domain . knowledge_level . abbreviation of
            unsupervised machine learning model + found cluster' i.e. C.2.Kme02

    """
    def __init__(self, knowledge_domain, knowledge_level, identifier,
                 learnblock_indices, learnblock_features, learnblock_labels,
                 t_min, t_max, subject, aim):
        assert isinstance(subject, tuple)

        self.knowledge_domain = knowledge_domain
        self.knowledge_level = knowledge_level
        self.identifier = identifier
        self.learnblock_indices = learnblock_indices
        self.learnblock_features = learnblock_features
        self.learnblock_labels = learnblock_labels
        self.t_min = t_min
        self.t_max = t_max
        self.subject = subject
        self.aim = aim
        self.subjects = None

    def __str__(self):
        return "\n".join([
            "{:20}:{}".format("Knowledge Domain", self.knowledge_domain),
            "{:20}:{}".format("Knowledge Level", self.knowledge_level),
            "{:20}:{}".format("Identifier", self.identifier),
            "{:20}:{}".format("Learnblock Indices", self.learnblock_indices),
            "{:20}:{}".format("Learnblock Labels", self.learnblock_labels),
            "{:20}:{}".format("Learnblock Features", self.learnblock_features),
            "{:20}:{}".format("T Min", self.t_min),
            "{:20}:{}".format("T Max", self.t_max),
            "{:20}:{}".format("Subject", self.subject),
            "{:20}:{}".format("Aim", self.aim)
            ]
        )

    def __repr__(self):
        return ", ".format([
                "{}={}".format("knowledge_domain", self.knowledge_domain),
                "{}={}".format("knowledge_level", self.knowledge_level),
                "{}={}".format("identifier", self.identifier),
                "{}={}".format("learnblock_indices", self.learnblock_indices),
                "{}={}".format("learnblock_features", self.learnblock_features),
                "{}={}".format("learnblock_labels", self.learnblock_features),
                "{}={}".format("t_min", self.t_min),
                "{}={}".format("t_max", self.t_max),
                "{}={}".format("subject", self.subject),
                "{}={}".format("aim", self.aim),
            ]
        )

    def __hash__(self):
        return hash(
            ".".join([
                self.knowledge_domain,
                str(self.knowledge_level),
                str(self.identifier)
            ])
        )

    def __eq__(self, other):
        if isinstance(other, Metadata):
            return hash(self) == hash(other)

        else:
            raise NotImplemented


class PragmaticMachineLearningModel:
    """Responsible of holding information about the pragmatic machine
    learning model.

    Args:
        meta (MetaData):
            Holds meta data of pragmatic machine laerning model.
        model (MachineLearningModel): Supervised machine learning model.
        learnblock (PandasBlock):
            Learnblock used for training the supervised machine learning model.

    """
    def __init__(self, meta, model, learnblock):
        if learnblock.constructed_from_models():
            self.__learnblock = learnblock
        else:
            self.__learnblock = None

        if meta.learnblock_labels is not None:
            assert meta.learnblock_labels[0] != ""

        self.meta = meta
        self.model = model
        self.origin = learnblock.origin

    def __str__(self):
        return "\n".join([
                str(self.meta),
                "{:20}: {}".format("Origin", str(self.origin))
            ]
        )

    def __repr__(self):
        return ", ".format([
                repr(self.meta),
                "{}={}".format("Origin", str(self.origin)),
                "{}=[{}]".format("model", repr(self.model))
            ]
        )

    def __hash__(self):
        return hash(self.meta)

    def __eq__(self, other):
        if isinstance(other, PragmaticMachineLearningModel):
            return hash(self.meta) == hash(other.meta)

        if isinstance(other, str):
            return hash(self) == hash(other)

        raise NotImplementedError

    @property
    def subjects(self):
        return self.meta.subjects

    @property
    def domain(self):
        return self.meta.knowledge_domain

    @property
    def level(self):
        return self.meta.knowledge_level

    @property
    def identifier(self):
        return self.meta.identifier

    @property
    def uid(self):
        return "." .join([self.domain, str(self.level), str(self.identifier)])

    @identifier.setter
    def identifier(self, value):
        self.meta.identifier = value

    @property
    def min_timestamp(self):
        return self.meta.t_min

    @property
    def max_timestamp(self):
        return self.meta.t_max

    @property
    def learnblock_features(self):
        return self.meta.learnblock_features

    @property
    def learnblock_indices(self):
        return self.meta.learnblock_indices

    @property
    def learnblock_labels(self):
        return self.meta.learnblock_labels

    @property
    def subject(self):
        return self.meta.subject

    @property
    def aim(self):
        return self.meta.aim

    @property
    def accuracy(self):
        return self.model.accuracy

    @property
    def reliability(self):
        return self.model.reliability

    @reliability.setter
    def reliability(self, value):
        self.model.reliability = value

    def fusion(self, model, knowledge_level, new_identifier):
        """Combine the meta data with the meta data  of another model.

        Args:
            model (PragmaticMachineLearningModel): With whome the fusion will
                kbe done.
            knowledge_level (int): Level on which the combination is performed.
            new_identifier (int): The new identifier for the combined model.

        Returns:
            MetaData:
                Combined medatdata.

        """
        meta_data = Metadata(knowledge_domain=self.meta.knowledge_domain,
                             knowledge_level=knowledge_level,
                             identifier=new_identifier,
                             learnblock_indices=None,
                             learnblock_features=None,
                             learnblock_labels=None,
                             t_min=min(self.meta.t_min, model.min_timestamp),
                             t_max=max(self.meta.t_max, model.max_timestamp),
                             subject=tuple(set(self.subject+model.subject)),
                             aim=tuple(set(self.meta.aim+model.aim)))
        meta_data.subjects = tuple(set(self.subjects + model.subjects))
        return meta_data

    def trained_with(self, knowledge):
        """Rebuild the learnblock on which the model was trained.

        Args:
            knowledge (KnowledgeDatabase): Database that stored the learnblockk.

        Returns:
            PandasBlock:
                Learnblock used for training the PragmaticMachineLearningModel.

        """
        if self.__learnblock:
            return self.__learnblock

        learnblock = knowledge.get_block(self.learnblock_indices)
        preserved_features = self.learnblock_features + ["T", "Sigma", "Z"]
        for feature in learnblock.columns():
            if feature not in preserved_features:
                learnblock.drop_columns([feature])
        learnblock.set_column("Z", self.learnblock_labels)

        learnblock.subject = self.subject
        learnblock.origin = self.origin
        return learnblock

    def set_subjects(self, subjects):
        """Assign to meta data all subjects.

        Args:
            subjects (list): List of abbreviations of supervised machine
                learning models.

        """
        self.meta.subjects = tuple(subjects)


class Reconstructor:
    """Responsible for reconstructing learnblocks.

    Notes:
        Only conceptual mode is currently supported.

    Args:
        mode (str): Either 'conceptual' or 'procedural'
        ml_models (list):
            List of supervised machine learning models.
        settings (ReconstrucionSettings):
            Instantiated object controlls the behavior of the Constructor.
        krippendorff_alpha (Callable):
            Wrapper for calculating interreability.

    Attributes:
        identifier_queue (Queue):
            Holds identifier for new reconstructed models.

    """
    CONCEPTUAL_KNOWLEDGE_ABBREVIATION = "C"
    PROCEDURAL_KNOWLEDGE_ABBREVIATION = "P"

    def __init__(self, mode, ml_models, settings, krippendorff_alpha):
        self.mode = mode
        self.ml_models = ml_models
        self.settings = settings
        self.identifier_queue = Manager().Queue()
        self.krippendorff_alpha = krippendorff_alpha

    def __str__(self):
        return "\n" .join([
            "{:20}: {}".format("Mode", str(self.mode)),
            "{:20}: {}".format("ML Models", [
                str(m) for m in self.ml_models]),
            str(self.settings)
        ]
        )

    def __repr__(self):
        return ", ".join([
            "{}={}".format("mode", str(self.mode)),
            "{}".format(repr(self.settings)),
            "{}={}".format("ml_models", repr(self.ml_models))
            ]
        )

    def select_winner(self, reliability_dict):
        return WinnerSelector().determine_winner(
            reliability_dict,
            self.ml_models
        )

    def reconstruct(self, level, lb, which_models, meta, *args, **kwargs):
        """Main method for reconstructing the given learnblock.

        Args:
            level (int): Level on which the reconstruction is performed.
            lb (PandasBlock): Learnblock.
            meta (Metadata): MetaData to use for the pragmatic machine learning
                model.

        Returns:
            tuple:
                Calculated inter-reliability as float.
                List with InterimPragmatics.
                List with logging information.

        """
        interims, logs = self.build_interim_pragmatics(
            level, lb, which_models, meta, *args, **kwargs
        )

        if len(interims) < self.settings.min_build_models:
            return 0.0, [], logs

        reliability = self.calc_intersubject_reliability(interims)

        for interim, log in zip_longest(interims, logs):
            if interim is not None:
                interim.model.reliability = reliability

            if log is not None:
                log.reliability = reliability

        if reliability < self.settings.min_reliability:
            return 0.0, [], logs

        reduced_interims = [i.model for i in interims]

        return reliability, reduced_interims, logs

    def build_interim_pragmatics(self,
                                 level,
                                 learnblock,
                                 which_models=None,
                                 meta=None,
                                 *args,
                                 **kwargs):
        """Train all supervised machine learning algorithms on the learnblock.


        Args:
            level (int): Level on which the reconstrucion is performed.
            learnblock (PandasBlock): Learnblock.
            which_models :
            meta (Metadata): MetaData to use for the pragmatic machine learning
                model.

        Returns:
            tuple:
                List with PragmaticMachineLearningModels.
                List with logging information.

        """
        train, test = split(learnblock, self.settings.reliability_sample)
        train_x = train.as_numpy_array()
        train_y = [label for label in train.get_column("Z")]
        test_x = test.as_numpy_array()
        test_y = [label for label in test.get_column("Z")]

        logs, models, subjects = [], [], []

        for model in self.machine_learning_models(which_models):
            trained_model = model.train(train_x, train_y)
            trained_model.accuracy = model.score(test_x, test_y)

            if valid_reconstructed(trained_model, self.mode, self.settings):
                predictions = trained_model.predict(test_x)
                metadata = self.build_meta(level, learnblock, trained_model, meta)
                pragmatic = PragmaticMachineLearningModel(metadata,
                                                          trained_model,
                                                          learnblock)

                models.append(InterimPragmatic(pragmatic, predictions))
                logs.append(log_info(
                    ReconstructionInfo.State.RECONSTRUCTED,
                    trained_model,
                    pragmatic))

                subjects.append(pragmatic.model.subject)

            else:
                logs.append(log_info(ReconstructionInfo.State.DISCARDED,
                                     trained_model))

        for m in models:
            m.model.set_subjects(subjects)

        return models, logs

    def machine_learning_models(self, which_models):
        """Yield unsupervised machine learning models.

        Args:
            which_models (list): List of unsupervised machine learning model
                abbreviations.

        Yields:
              MachineLearningModel

        """
        if not which_models:
            for model in self.ml_models:
                yield model
        else:
            for model in self.ml_models:
                if model.subject in which_models:
                    yield model

    def calc_intersubject_reliability(self, interims):
        """Calculate the interreabililty between labels of unsupervised
        machine learning models.

        Args:
            interims (list): List with InterimPragmatics

        Returns:
            float:
                Inter-reliability value between labels.

        """
        predictions = [interim.predictions for interim in interims]
        if len(predictions) == 1:
            return 0.0

        reliability_alpha = self.krippendorff_alpha(predictions)
        return reliability_alpha

    def build_meta(self, level, learnblock, model, meta):
        """Build meta data for the reconstructed model.

        Args:
            level (int): Level on which the reconstrucion is performed.
            learnblock (PandasBlock): Learnblock.
            model (MachineLearningModel): Supervised machine learnign model.
            meta (Metadata): Use it a basis for the new meta data.

        Returns:
            MetaData

        """
        if not meta:
            zeta = self.build_aim(level, learnblock)
            identifier = self.identifier_queue.get()

            return Metadata(
                knowledge_domain=self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION,
                knowledge_level=level,
                identifier=identifier,
                learnblock_indices=learnblock.indices(),
                learnblock_features=learnblock.columns(),
                learnblock_labels=learnblock.get_column("Z"),
                t_min=learnblock.min_timestamp(),
                t_max=learnblock.max_timestamp(),
                subject=(model.subject,),
                aim=(zeta, ))

        else:
            meta.learnblock_indices = learnblock.indices()
            meta.learnblock_features = learnblock.columns()
            meta.learnblock_labels = learnblock.get_column("Z")
            return meta

    def build_aim(self, level, learnblock):
        """Build the aim of the new reconstruced learnblock.

        Args:
            level (int): Knwoledge domain level.
            learnblock (PandasBlock): Used learnblock.

        Returns:
            str:
                Aim for PragmaticMachineLearningModel: C.2.Kme02

        """
        return ".".join([self.CONCEPTUAL_KNOWLEDGE_ABBREVIATION, str(level),
                         str(learnblock.subject)])

    def get_pml(self, meta, block, ml_model):
        """Build PragmaticMachineLearningModel without reconstrucing
        it but with just the given meta data.

        Args:
            meta (Metadata): Meta data.
            block (PandasBlock): Learnblock.
            ml_model (List): List of supervised machine learning models.

        Returns:
            PragmaticMachineLearningModels

        """
        meta.learnblock_indicies = block.indices()
        meta.learnblock_features = block.columns()
        meta.learnblock_labels = block.get_column("Z")
        return PragmaticMachineLearningModel(meta, ml_model, block)

    @property
    def min_test_accuracy(self):
        return self.settings.min_test_accuracy

    @min_test_accuracy.setter
    def min_test_accuracy(self, value):
        self.settings.min_test_accuracy = value

    @property
    def reliability_sample(self):
        return self.settings.reliability_sample

    @reliability_sample.setter
    def reliability_sample(self, value):
        self.settings.reliability_sample = value

    @property
    def min_reliability(self):
        return self.settings.min_reliability

    @min_reliability.setter
    def min_reliability(self, value):
        self.settings.min_reliability = value


def split(learnblock, reliability_sample):
    indices = learnblock.indices()
    eval_size = int(learnblock.n_rows() * reliability_sample)
    eval_idx, train_idx = indices[:eval_size], indices[eval_size:]
    return (
        learnblock.new_block_from(train_idx, index=True),
        learnblock.new_block_from(eval_idx, index=True)
    )


def valid_reconstructed(model, knowledge_domain, settings):
    """Test if the trained supersived model fullfills the accuracy criteria.

    Args:
        model (MachineLearningModel):
        knowledge_domain (str): Either 'conceptual' or 'procedural'.
        settings (ReconstructionSettings): Contains the minimum accuracy.

    Returns:
        bool:
            True if accuracy criteria is fullfilled else False.

    """
    if knowledge_domain == "conceptual":
        return model.accuracy >= settings.min_test_accuracy


def log_info(state, trained_model, pragmatic=None):
    """Build logging object for the reconstruction process.

    Args:
        state (Enum): Indicates the state of the reconstruction.
            If it was successfull or not.
        trained_model (MachineLearningModel): Supervised model.

    Returns:
        ReconstructionInfo
            Holds information about the finished reconstruction process.

    """
    uid = pragmatic.uid if pragmatic else ""
    aim = pragmatic.aim if pragmatic else ""

    return ReconstructionInfo(
        state=state,
        uid=uid,
        aim=aim,
        accuracy=trained_model.accuracy,
        subject=trained_model.subject
    )
